"""Test exceptions."""

import pytest

from cx_Freeze.exception import ConfigError, DarwinException


@pytest.mark.parametrize("custom_exception", [ConfigError, DarwinException])
def test_raise_exceptions(custom_exception):
    """This method tests that exceptions can be raised + caught and the __str__
    value is correctly called"""
    try:
        raise custom_exception("Something Bad Happened")
    except custom_exception as err:
        print(str(err))  # Force invokation of the __str__ method



"""A collection of functions which are triggered automatically by finder when
pycryptodome package is included."""

import os

from ..common import code_object_replace
from ..finder import ModuleFinder
from ..module import Module


def load_crypto_cipher(finder: ModuleFinder, module: Module) -> None:
    """The Crypto.Cipher subpackage of pycryptodome package."""
    if module.in_file_system == 0:
        finder.include_package(module.name)


def load_crypto_hash(finder: ModuleFinder, module: Module) -> None:
    """The Crypto.Hash subpackage of pycryptodome package."""
    if module.in_file_system == 0:
        finder.include_package(module.name)


def load_crypto_math(finder: ModuleFinder, module: Module) -> None:
    """The Crypto.Math subpackage of pycryptodome package."""
    if module.in_file_system == 0:
        finder.include_package(module.name)


def load_crypto_protocol(finder: ModuleFinder, module: Module) -> None:
    """The Crypto.Protocol subpackage of pycryptodome package."""
    if module.in_file_system == 0:
        finder.include_package(module.name)


def load_crypto_publickey(finder: ModuleFinder, module: Module) -> None:
    """The Crypto.PublicKey subpackage of pycryptodome package."""
    if module.in_file_system == 0:
        finder.include_package(module.name)


def load_crypto_util(finder: ModuleFinder, module: Module) -> None:
    """The Crypto.Util subpackage of pycryptodome package."""
    if module.in_file_system == 0:
        finder.include_package(module.name)


def load_crypto_util__file_system(
    finder: ModuleFinder, module: Module  # pylint: disable=unused-argument
) -> None:
    """The patch for pycryptodome package."""
    # WARNING: do not touch this code string
    code_to_inject = """
import os

def pycryptodome_filename(dir_comps, filename):
    import sys
    if dir_comps[0] != "Crypto":
        raise ValueError("Only available for modules under 'Crypto'")
    dir_comps = list(dir_comps) + [filename]
    root_lib = os.path.join(os.path.dirname(sys.executable), "lib")
    return os.path.join(root_lib, ".".join(dir_comps))
"""
    if module.in_file_system == 0 and module.code is not None:
        new_code = compile(code_to_inject, os.fspath(module.file), "exec")
        co_func = new_code.co_consts[2]
        name = co_func.co_name
        code = module.code
        consts = list(code.co_consts)
        for i, constant in enumerate(consts):
            if isinstance(constant, type(code)) and constant.co_name == name:
                consts[i] = co_func
                break
        module.code = code_object_replace(code, co_consts=consts)



"""Shared fixtures"""

import os
from pathlib import Path

import pytest


# pylint: disable=redefined-outer-name
@pytest.fixture()
def fix_test_path() -> Path:
    """This fixture returns the root of the test folder"""
    return Path(__file__).resolve().parent


@pytest.fixture()
def fix_test_samples_path(fix_test_path: Path) -> Path:
    """This fixture returns the samples folder for the tests"""
    return fix_test_path / "samples"


@pytest.fixture()
def fix_test_samples_dir(fix_test_samples_path: Path) -> str:
    """This fixture returns the samples folder for the tests"""
    return os.fspath(fix_test_samples_path)


@pytest.fixture()
def fix_main_samples_path(fix_test_path: Path) -> Path:
    """This fixture returns the cx_Freeze samples folder"""
    return fix_test_path.parent.joinpath("samples").resolve()



"""Extend setuptools.sandbox.run_setup to work with cx_Freeze setup()."""

import contextlib
import os
import sys
from pathlib import Path
from typing import Sequence, Union

import setuptools.sandbox as _sandbox

# pylint: disable-next=protected-access
_sandbox._MODULES_TO_HIDE.update({"cx_Freeze", "msilib"})

__all__ = ["run_setup"]


def run_setup(setup_script: Union[str, Path], args: Sequence):
    """Run a cx_Freeze setup script, sandboxed in its directory."""

    if isinstance(setup_script, Path):
        setup_script = os.fspath(setup_script)
    setup_dir = os.path.abspath(os.path.dirname(setup_script))
    with _setup_context(setup_dir):
        try:
            sys.argv[:] = [setup_script] + list(args)
            sys.path.insert(0, setup_dir)
            with _sandbox.DirectorySandbox(setup_dir):
                env = {"__file__": setup_script, "__name__": "__main__"}
                # pylint: disable-next=protected-access
                _sandbox._execfile(setup_script, env)
        except SystemExit as exc:
            if exc.args and exc.args[0]:
                raise
            # Normal exit, just return


@contextlib.contextmanager
def _setup_context(setup_dir: str):
    """Save and prepare de environment to run_setup."""

    with _sandbox.setup_context(setup_dir):
        # ensure cx_Freeze commands are available
        __import__("cx_Freeze")
        yield



"""
A simple setup script to create an executable using PyQt6-WebEngine.
This also demonstrates how to use excludes to get minimal package size.

Run the build process by running the command 'python setup.py build'

If everything works well you should find a subdirectory in the build
subdirectory that contains the files needed to run the application
"""

import sys

from cx_Freeze import Executable, setup

try:
    from cx_Freeze.hooks import get_qt_plugins_paths
except ImportError:
    get_qt_plugins_paths = None

include_files = []
if get_qt_plugins_paths:
    # Inclusion of extra plugins (since cx_Freeze 6.8b2)
    # cx_Freeze imports automatically the following plugins depending of the
    # use of some modules:
    # imageformats, platforms, platformthemes, styles - QtGui
    # mediaservice - QtMultimedia
    # printsupport - QtPrintSupport
    for plugin_name in (
        # "accessible",
        # "iconengines",
        # "platforminputcontexts",
        # "xcbglintegrations",
        # "egldeviceintegrations",
        "wayland-decoration-client",
        "wayland-graphics-integration-client",
        # "wayland-graphics-integration-server",
        "wayland-shell-integration",
    ):
        include_files += get_qt_plugins_paths("PyQt6", plugin_name)

base = "Win32GUI" if sys.platform == "win32" else None

build_exe_options = {
    "bin_excludes": ["libqpdf.so", "libqpdf.dylib"],
    # exclude packages that are not really needed
    "excludes": ["tkinter", "unittest", "email", "http", "xml", "pydoc"],
    "include_files": include_files,
    "zip_include_packages": ["PyQt6"],
}

executables = [
    Executable("simplebrowser.py", target_name="test_simplebrowser")
]

setup(
    name="simplebrowser",
    version="0.1",
    description="Sample cx_Freeze PyQt6 simplebrowser script",
    options={"build_exe": build_exe_options},
    executables=executables,
)



"""Tests for cx_Freeze.command.build."""

import os
import sys
from sysconfig import get_platform, get_python_version

from cx_Freeze.command.build import Build
from cx_Freeze.dist import Distribution


# pylint: disable=C0115,C0116
class TestBuild:
    def test_finalize_options(self):
        dist = Distribution(
            {"name": "foo", "version": "0.0", "script_name": "setup.py"}
        )
        cmd = Build(dist)
        cmd.finalize_options()

        # if not specified, plat_name gets the current platform
        assert cmd.plat_name == get_platform()

        # build_purelib is build + lib
        wanted = os.path.join(cmd.build_base, "lib")
        assert cmd.build_purelib == wanted

        # build_platlib is 'build/lib.platform-x.x[-pydebug]'
        # examples:
        #   build/lib.linux-x86_64-3.9
        #   build/lib.win-amd64-3.10
        python_version = get_python_version()
        plat_spec = f".{cmd.plat_name}-{python_version}"
        if hasattr(sys, "gettotalrefcount"):
            self.assertTrue(cmd.build_platlib.endswith("-pydebug"))
            plat_spec += "-pydebug"
        wanted = os.path.join(cmd.build_base, "lib" + plat_spec)
        assert cmd.build_platlib == wanted

        # by default, build_lib = build_purelib
        assert cmd.build_lib == cmd.build_purelib

        # build_temp is build/temp.<plat>
        wanted = os.path.join(cmd.build_base, "temp" + plat_spec)
        assert cmd.build_temp == wanted

        # build_scripts is build/scripts-x.x
        wanted = os.path.join(cmd.build_base, f"scripts-{python_version}")
        assert cmd.build_scripts == wanted

        # executable is os.path.normpath(sys.executable)
        assert cmd.executable == os.path.normpath(sys.executable)



"""A collection of functions which are triggered automatically by finder when
scipy package is included."""
# pylint: disable=unused-argument

import sys

from ..finder import ModuleFinder
from ..module import Module

WIN32 = sys.platform == "win32"


def load_scipy(finder: ModuleFinder, module: Module) -> None:
    """The scipy module loads items within itself in a way that causes
    problems without libs and a number of subpackages being present."""
    libs_name = f"{module.name}.libs"
    source_dir = module.path[0].parent / libs_name
    if source_dir.exists():
        finder.include_files(source_dir, f"lib/{libs_name}")

    finder.include_package("scipy.integrate")
    finder.include_package("scipy._lib")
    finder.include_package("scipy.misc")
    finder.include_package("scipy.optimize")
    if WIN32:
        finder.exclude_module("scipy.spatial.cKDTree")


def load_scipy_linalg(finder: ModuleFinder, module: Module) -> None:
    """The scipy.linalg module loads items within itself in a way that causes
    problems without the entire package being present."""
    module.global_names.add("norm")
    finder.include_package("scipy.linalg")


def load_scipy_linalg_interface_gen(
    finder: ModuleFinder, module: Module
) -> None:
    """The scipy.linalg.interface_gen module optionally imports the pre module;
    ignore the error if this module cannot be found."""
    module.ignore_names.add("pre")


def load_scipy_ndimage(finder: ModuleFinder, module: Module) -> None:
    """The scipy.ndimage must be loaded as a package."""
    finder.exclude_module("scipy.ndimage.tests")
    finder.include_package("scipy.ndimage")


def load_scipy_sparse_csgraph(finder: ModuleFinder, module: Module) -> None:
    """The scipy.sparse.csgraph must be loaded as a package."""
    finder.exclude_module("scipy.sparse.csgraph.tests")
    finder.include_package("scipy.sparse.csgraph")


def load_scipy_sparse_linalg_dsolve_linsolve(
    finder: ModuleFinder, module: Module
) -> None:
    """The scipy.linalg.dsolve.linsolve optionally loads scikits.umfpack."""
    module.ignore_names.add("scikits.umfpack")


def load_scipy_spatial_transform(finder: ModuleFinder, module: Module) -> None:
    """The scipy.spatial.transform must be loaded as a package."""
    finder.include_package("scipy.spatial.transform")
    finder.exclude_module("scipy.spatial.transform.tests")


def load_scipy_special(finder: ModuleFinder, module: Module) -> None:
    """The scipy.special must be loaded as a package."""
    finder.include_package("scipy.special")


def load_scipy_special__cephes(finder: ModuleFinder, module: Module) -> None:
    """The scipy.special._cephes is an extension module and the scipy module
    imports * from it in places; advertise the global names that are used
    in order to avoid spurious errors about missing modules."""
    module.global_names.add("gammaln")


def load_scipy_stats(finder: ModuleFinder, module: Module) -> None:
    """The scipy.stats must be loaded as a package."""
    finder.include_package("scipy.stats")
    finder.exclude_module("scipy.stats.tests")



"""Tests for cx_Freeze.command.bdist_rpm."""

import sys
from pathlib import Path
from sysconfig import get_platform

import pytest

from cx_Freeze.sandbox import run_setup


@pytest.mark.skipif(sys.platform != "linux", reason="Linux tests")
def test_bdist_rpm(fix_main_samples_path: Path):
    """Test the bcrypt sample with bdist_rpm."""

    package = "bcrypt"
    version = "0.3"
    arch = get_platform().split("-", 1)[1]
    setup_path = fix_main_samples_path / package
    dist_created = setup_path / "dist"
    dist_already_exists = dist_created.exists()

    run_setup(setup_path / "setup.py", ["bdist_rpm"])

    base_name = f"test_{package}-{version}"

    file_created = dist_created / f"{base_name}-1.src.rpm"
    assert file_created.is_file()
    file_created.unlink()

    file_created = dist_created / f"{base_name}-1.{arch}.rpm"
    assert file_created.is_file()
    file_created.unlink()

    file_created = dist_created / f"{base_name}.tar.gz"
    assert file_created.is_file()
    file_created.unlink()

    if not dist_already_exists:
        dist_created.rmdir()



"""A collection of functions which are triggered automatically by finder when
numpy package is included."""
# pylint: disable=unused-argument

import os
import sys
from importlib.machinery import EXTENSION_SUFFIXES
from pathlib import Path

from ..finder import ModuleFinder
from ..module import Module

WIN32 = sys.platform == "win32"


def load_numpy(finder: ModuleFinder, module: Module) -> None:
    """The numpy must be loaded as a package; support for pypi version and
    numpy+mkl version - tested with 1.19.5+mkl, 1.20.3+mkl, 1.21.0+mkl,
    1.21.1+mkl, 1.21.2+mkl and 1.21.2 from conda-forge."""
    finder.include_package("numpy")

    if WIN32:
        numpy_dir = module.path[0]
        # numpy+mkl from: https://www.lfd.uci.edu/~gohlke/pythonlibs/#numpy
        libs_dir = numpy_dir / "DLLs"
        if not libs_dir.is_dir():
            # numpy+mkl from conda-forge
            libs_dir = Path(sys.base_prefix, "Library", "bin")
        if libs_dir.is_dir():
            dest_dir = Path("lib", "numpy_mkl")
            for path in libs_dir.glob("mkl_*.dll"):
                finder.include_files(path, dest_dir / path.name)
            for path in libs_dir.glob("lib*.dll"):
                finder.include_files(path, dest_dir / path.name)
            finder.add_constant("MKL_PATH", os.fspath(dest_dir))
            finder.exclude_module("numpy.DLLs")

            # do not check dependencies already handled
            extension = EXTENSION_SUFFIXES[0]
            for path in numpy_dir.rglob(f"*{extension}"):
                finder.exclude_dependent_files(path)

        # support for old versions (numpy <= 1.18.2)
        if module.in_file_system == 0:
            # copy any file at site-packages/numpy/.libs
            libs_dir = numpy_dir / ".libs"
            if libs_dir.is_dir():
                finder.include_files(libs_dir, "lib")

    # exclude the tests
    finder.exclude_module("numpy.compat.tests")
    finder.exclude_module("numpy.core.tests")
    finder.exclude_module("numpy.distutils.tests")
    finder.exclude_module("numpy.f2py.tests")
    finder.exclude_module("numpy.fft.tests")
    finder.exclude_module("numpy.lib.tests")
    finder.exclude_module("numpy.linalg.tests")
    finder.exclude_module("numpy.ma.tests")
    finder.exclude_module("numpy.matrixlib.tests")
    finder.exclude_module("numpy.polynomial.tests")
    finder.exclude_module("numpy.random._examples")
    finder.exclude_module("numpy.random.tests")
    finder.exclude_module("numpy.tests")
    finder.exclude_module("numpy.typing.tests")


def load_numpy_core_numerictypes(finder: ModuleFinder, module: Module) -> None:
    """The numpy.core.numerictypes module adds a number of items to itself
    dynamically; define these to avoid spurious errors about missing
    modules."""
    module.global_names.update(
        [
            "bool_",
            "cdouble",
            "complexfloating",
            "csingle",
            "double",
            "float64",
            "float_",
            "inexact",
            "intc",
            "int32",
            "number",
            "single",
        ]
    )


def load_numpy_distutils_command_scons(
    finder: ModuleFinder, module: Module
) -> None:
    """The numpy.distutils.command.scons module optionally imports the numscons
    module; ignore the error if the module cannot be found."""
    module.ignore_names.add("numscons")


def load_numpy_distutils_misc_util(
    finder: ModuleFinder, module: Module
) -> None:
    """The numpy.distutils.misc_util module optionally imports the numscons
    module; ignore the error if the module cannot be found."""
    module.ignore_names.add("numscons")


def load_numpy_distutils_system_info(
    finder: ModuleFinder, module: Module
) -> None:
    """The numpy.distutils.system_info module optionally imports the Numeric
    module; ignore the error if the module cannot be found."""
    module.ignore_names.add("Numeric")


def load_numpy_f2py___version__(finder: ModuleFinder, module: Module) -> None:
    """The numpy.f2py.__version__ module optionally imports the __svn_version__
    module; ignore the error if the module cannot be found."""
    module.ignore_names.add("__svn_version__")


def load_numpy_linalg(finder: ModuleFinder, module: Module) -> None:
    """The numpy.linalg module implicitly loads the lapack_lite module; make
    sure this happens."""
    finder.include_module("numpy.linalg.lapack_lite")


def load_numpy_random_mtrand(finder: ModuleFinder, module: Module) -> None:
    """The numpy.random.mtrand module is an extension module and the numpy
    module imports * from this module; define the list of global names
    available to this module in order to avoid spurious errors about missing
    modules."""
    module.global_names.update(["rand", "randn"])



"""Code to test that the plist_items option is working correctly."""

import plistlib
import sys

import pytest
from data import BUILD_DIR, BUNDLE_NAME, TEST_KEY, TEST_VALUE

from cx_Freeze.sandbox import run_setup


@pytest.mark.skipif(sys.platform != "darwin", reason="Macos tests")
def test_plist_items(fix_test_samples_path):
    """Test that the plist_items option is working correctly."""

    setup_path = fix_test_samples_path / "plist_items"

    run_setup(setup_path / "setup.py", ["bdist_mac"])

    # Test that the additional keys were correctly added to the plist.
    path = setup_path / f"{BUILD_DIR}/{BUNDLE_NAME}.app/Contents/Info.plist"
    contents = plistlib.loads(path.read_bytes())
    assert contents[TEST_KEY] == TEST_VALUE



# pylint: skip-file
import moda  # noqa
from modb import b  # noqa

from . import modc  # noqa
from .modd import d  # noqa

from mode import *  # noqa isort:skip
from ..modf import *  # noqa isort:skip
import modg.submod  # noqa isort:skip

try:
    pass
finally:
    import modh  # noqa



"""Tests for cx_Freeze.command.bdist_msi."""
import sys
from pathlib import Path
from sysconfig import get_platform

import pytest

from cx_Freeze.dist import Distribution
from cx_Freeze.sandbox import run_setup

if sys.platform == "win32":
    from cx_Freeze.command.bdist_msi import BdistMSI


@pytest.mark.skipif(sys.platform != "win32", reason="Windows tests")
@pytest.mark.skipif(
    sys.version_info < (3, 7, 4), reason="requires python 3.7.4 or higher"
)
def test_bdist_msi(fix_main_samples_path: Path):
    """Test the msi_binary_data sample."""

    setup_path = fix_main_samples_path / "msi_binary_data"
    dist_created = setup_path / "dist"
    dist_already_exists = dist_created.exists()

    run_setup(setup_path / "setup.py", ["bdist_msi"])

    platform = get_platform().replace("win-amd64", "win64")
    file_created = dist_created / f"hello-0.1-{platform}.msi"
    assert file_created.is_file()
    file_created.unlink()

    if not dist_already_exists:
        dist_created.rmdir()


@pytest.mark.skipif(sys.platform != "win32", reason="Windows tests")
def test_bdist_msi_target_name():
    """Test the bdist_msi with extra target_name option."""

    dist = Distribution(
        {"name": "foo", "version": "0.0", "script_name": "setup.py"}
    )
    cmd = BdistMSI(dist)
    cmd.target_name = "mytest"
    cmd.finalize_options()
    cmd.ensure_finalized()

    assert cmd.fullname == "mytest-0.0"


@pytest.mark.skipif(sys.platform != "win32", reason="Windows tests")
def test_bdist_msi_target_name_and_version():
    """Test the bdist_msi with extra target options."""

    dist = Distribution(
        {"name": "foo", "version": "0.0", "script_name": "setup.py"}
    )
    cmd = BdistMSI(dist)
    cmd.target_name = "mytest"
    cmd.target_version = "0.1"
    cmd.finalize_options()
    cmd.ensure_finalized()
    assert cmd.fullname == "mytest-0.1"


@pytest.mark.skipif(sys.platform != "win32", reason="Windows tests")
@pytest.mark.skipif(
    sys.version_info < (3, 7, 4), reason="requires python 3.7.4 or higher"
)
def test_bdist_msi_target_name_with_extension(fix_main_samples_path: Path):
    """Test the msi_binary_data sample, with a specified target_name that
    includes an ".msi" extension."""

    msi_name = "output.msi"
    setup_path = fix_main_samples_path / "msi_binary_data"
    dist_created = setup_path / "dist"
    dist_already_exists = dist_created.exists()

    run_setup(
        setup_path / "setup.py", ["bdist_msi", "--target-name", msi_name]
    )

    file_created = dist_created / msi_name
    assert file_created.is_file()
    file_created.unlink()

    if not dist_already_exists:
        dist_created.rmdir()



"""
Implements a simple service using cx_Freeze.

See below for more information on what methods must be implemented and how they
are called.
"""

import asyncio
import threading


class Handler:

    # no parameters are permitted; all configuration should be placed in the
    # configuration file and handled in the initialize() method
    def __init__(self):
        # The loop MUST be created here.
        self.loop = asyncio.ProactorEventLoop()

        self.stopEvent = threading.Event()
        self.stopRequestedEvent = asyncio.Event()

    # called when the service is starting
    def initialize(self, configFileName):
        pass

    # called when the service is starting immediately after initialize()
    # use this to perform the work of the service; don't forget to set or check
    # for the stop event or the service GUI will not respond to requests to
    # stop the service
    def run(self):
        self.loop.run_until_complete(self.stopRequestedEvent.wait())
        self.stopEvent.set()

    # called when the service is being stopped by the service manager GUI
    def stop(self):
        self.loop.call_soon_threadsafe(self.stopRequestedEvent.set)
        self.stopEvent.wait()
